@@ -1,10 +1,11 @@
 #include "main.h"
 
-#include "camera_config.h"
 #include "flash.h"
 #include "gl_util.h"
 #include "io_pipeline.h"
 #include "process_pipeline.h"
+#include "rotation.h"
+#include "state.h"
 #include <asm/errno.h>
 #include <assert.h>
 #include <errno.h>
@@ -21,6 +22,7 @@
 #include <X11/extensions/Xrandr.h>
 #include <gdk/x11/gdkx.h>
 #endif
+#include <libmegapixels.h>
 #include <limits.h>
 #include <linux/kdev_t.h>
 #include <linux/media.h>
@@ -43,42 +45,18 @@
 RENDERDOC_API_1_1_2 *rdoc_api = NULL;
 #endif
 
-#define APP_ID "org.postmarketos.Megapixels"
+mp_state_main state;
 
-enum user_control { USER_CONTROL_ISO, USER_CONTROL_SHUTTER };
-
-static bool camera_is_initialized = false;
-static const struct mp_camera_config *camera = NULL;
-static MPMode mode;
-
-static int preview_width = -1;
-static int preview_height = -1;
-
-static int device_rotation = 0;
-
-static bool gain_is_manual = false;
-static int gain;
-static int gain_max;
-
-static bool exposure_is_manual = false;
-static int exposure;
-
-static bool has_auto_focus_continuous;
-static bool has_auto_focus_start;
-
-static bool flash_enabled = false;
+static int movie_start = 0;
 
 static MPProcessPipelineBuffer *current_preview_buffer = NULL;
-static int preview_buffer_width = -1;
-static int preview_buffer_height = -1;
 
 static char last_path[260] = "";
 
 static MPZBarScanResult *zbar_result = NULL;
 
-static int burst_length = 0;
-
 // Widgets
+GtkWidget *window;
 GtkWidget *preview;
 GtkWidget *main_stack;
 GtkWidget *open_last_stack;
@@ -87,8 +65,18 @@ GtkWidget *process_spinner;
 GtkWidget *scanned_codes;
 GtkWidget *preview_top_box;
 GtkWidget *preview_bottom_box;
+
+GtkWidget *message_box;
+GtkWidget *message_label;
+
+GtkWidget *iso_button;
+GtkWidget *shutter_button;
+GtkWidget *focus_button;
 GtkWidget *flash_button;
+GtkWidget *status_button;
+
 LfbEvent *capture_event;
+static GtkWidget *movie;
 
 GSettings *settings;
 GSettings *fb_settings;
@@ -111,60 +99,161 @@ remap(int value, int input_min, int input_max, int output_min, int output_max)
 }
 
 static void
-update_io_pipeline()
+display_error(const char *message)
 {
-        struct mp_io_pipeline_state io_state = {
-                .camera = camera,
-                .burst_length = burst_length,
-                .preview_width = preview_width,
-                .preview_height = preview_height,
-                .device_rotation = device_rotation,
-                .gain_is_manual = gain_is_manual,
-                .gain = gain,
-                .exposure_is_manual = exposure_is_manual,
-                .exposure = exposure,
-                .flash_enabled = flash_enabled,
-        };
-        mp_io_pipeline_update_state(&io_state);
-
-        // Make the right settings available for the camera
-        gtk_widget_set_visible(flash_button, camera->has_flash);
+        gtk_label_set_label(GTK_LABEL(message_label), message);
+        gtk_widget_set_visible(message_box, true);
 }
 
-static bool
-update_state(const struct mp_main_state *state)
+bool
+check_window_active()
 {
-        if (!camera_is_initialized) {
-                camera_is_initialized = true;
+        return gtk_window_is_active(GTK_WINDOW(window));
+}
+
+static void
+update_io_pipeline()
+{
+        mp_state_io new_state = {
+                .camera = state.camera,
+                .configuration = state.configuration,
+                .burst_length = state.burst_length,
+                .preview_width = state.preview_width,
+                .preview_height = state.preview_height,
+                .device_rotation = state.device_rotation,
+                .device_accel_rotation = state.device_accel_rotation,
+                .device_accel_rotation_good = state.device_accel_rotation_good,
+
+                .gain.control = state.gain.control,
+                .gain.auto_control = state.gain.auto_control,
+                .gain.value = state.gain.value,
+                .gain.value_req = state.gain.value_req,
+                .gain.max = state.gain.max,
+                .gain.min = state.gain.min,
+                .gain.manual = state.gain.manual,
+                .gain.manual_req = state.gain.manual_req,
+
+                .dgain.control = state.dgain.control,
+                .dgain.auto_control = state.dgain.auto_control,
+                .dgain.value = state.dgain.value,
+                .dgain.value_req = state.dgain.value_req,
+                .dgain.max = state.dgain.max,
+                .dgain.min = state.dgain.min,
+                .dgain.manual = state.dgain.manual,
+                .dgain.manual_req = state.dgain.manual_req,
+
+                .exposure.control = state.exposure.control,
+                .exposure.auto_control = state.exposure.auto_control,
+                .exposure.value = state.exposure.value,
+                .exposure.value_req = state.exposure.value_req,
+                .exposure.max = state.exposure.max,
+                .exposure.manual = state.exposure.manual,
+                .exposure.manual_req = state.exposure.manual_req,
+
+                .focus.control = state.focus.control,
+                .focus.auto_control = state.focus.auto_control,
+                .focus.value = state.focus.value,
+                .focus.value_req = state.focus.value_req,
+                .focus.max = state.focus.max,
+                .focus.manual = state.focus.manual,
+                .focus.manual_req = state.focus.manual_req,
+
+                .stats.exposure = state.stats.exposure,
+                .stats.temp = state.stats.temp,
+                .stats.tint = state.stats.tint,
+                .stats.focus = state.stats.focus,
+
+                .flash_enabled = state.flash_enabled,
+        };
+        mp_io_pipeline_update_state(&new_state);
+}
+
+void
+notify_auto_status(gchar *msg)
+{
+	gtk_button_set_label(GTK_BUTTON(status_button), msg);
+}
+
+/*
+ * State transfer from Process -> Main
+ */
+static bool
+update_state(const mp_state_main *new_state)
+{
+        if (state.exposure.manual != new_state->exposure.manual) {
+                const char *icon_name = new_state->exposure.manual ?
+                                                "shutter-man-symbolic" :
+                                                "shutter-auto-symbolic";
+                gtk_button_set_icon_name(GTK_BUTTON(shutter_button), icon_name);
         }
 
-        if (camera == state->camera) {
-                mode = state->mode;
-
-                if (!gain_is_manual) {
-                        gain = state->gain;
-                }
-                gain_max = state->gain_max;
-
-                if (!exposure_is_manual) {
-                        exposure = state->exposure;
-                }
-
-                has_auto_focus_continuous = state->has_auto_focus_continuous;
-                has_auto_focus_start = state->has_auto_focus_start;
+        if (state.gain.manual != new_state->gain.manual) {
+                const char *icon_name = new_state->gain.manual ? "iso-man-symbolic" :
+                                                                 "iso-auto-symbolic";
+                gtk_button_set_icon_name(GTK_BUTTON(iso_button), icon_name);
         }
 
-        preview_buffer_width = state->image_width;
-        preview_buffer_height = state->image_height;
+        if (state.focus.manual != new_state->focus.manual) {
+                const char *icon_name = new_state->focus.manual ?
+                                                "focus-man-symbolic" :
+                                                "focus-auto-symbolic";
+                gtk_button_set_icon_name(GTK_BUTTON(focus_button), icon_name);
+        }
+
+        state.gain.control = new_state->gain.control;
+        state.gain.auto_control = new_state->gain.auto_control;
+        state.gain.value = new_state->gain.value;
+        state.gain.max = new_state->gain.max;
+        state.gain.min = new_state->gain.min;
+        state.gain.manual = new_state->gain.manual;
+
+        state.dgain.control = new_state->dgain.control;
+        state.dgain.auto_control = new_state->dgain.auto_control;
+        state.dgain.value = new_state->dgain.value;
+        state.dgain.max = new_state->dgain.max;
+        state.dgain.min = new_state->dgain.min;
+        state.dgain.manual = new_state->dgain.manual;
+
+        state.exposure.control = new_state->exposure.control;
+        state.exposure.auto_control = new_state->exposure.auto_control;
+        state.exposure.value = new_state->exposure.value;
+        state.exposure.max = new_state->exposure.max;
+        state.exposure.manual = new_state->exposure.manual;
+
+        state.focus.control = new_state->focus.control;
+        state.focus.auto_control = new_state->focus.auto_control;
+        state.focus.value = new_state->focus.value;
+        state.focus.max = new_state->focus.max;
+        state.focus.manual = new_state->focus.manual;
+
+        state.has_auto_focus_continuous = new_state->has_auto_focus_continuous;
+        state.has_auto_focus_start = new_state->has_auto_focus_start;
+
+        state.preview_buffer_width = new_state->preview_buffer_width;
+        state.preview_buffer_height = new_state->preview_buffer_height;
+
+        state.stats.exposure = new_state->stats.exposure;
+        state.stats.temp = new_state->stats.temp;
+        state.stats.tint = new_state->stats.tint;
+        state.stats.focus = new_state->stats.focus;
+
+        // Make the right settings available for the camera
+        gtk_widget_set_visible(iso_button, state.gain.control.id != 0);
+        gtk_widget_set_visible(shutter_button, state.exposure.control.id != 0);
+        gtk_widget_set_visible(focus_button, state.focus.control.id != 0);
+        // Even if there's no flash led/v4l, it'll just default to using the screen
+        // as flash, so always enable this button
+        gtk_widget_set_visible(flash_button, true);
+        gtk_widget_set_visible(status_button, true);
 
         return false;
 }
 
 void
-mp_main_update_state(const struct mp_main_state *state)
+mp_main_update_state(const mp_state_main *new_state)
 {
-        struct mp_main_state *state_copy = malloc(sizeof(struct mp_main_state));
-        *state_copy = *state;
+        mp_state_main *state_copy = malloc(sizeof(mp_state_main));
+        *state_copy = *new_state;
 
         g_main_context_invoke_full(g_main_context_default(),
                                    G_PRIORITY_DEFAULT_IDLE,
@@ -229,7 +318,7 @@ struct capture_completed_args {
 static bool
 capture_completed(struct capture_completed_args *args)
 {
-        strncpy(last_path, args->fname, 259);
+        snprintf(last_path, sizeof(last_path), "%s", args->fname);
 
         gtk_image_set_from_paintable(GTK_IMAGE(thumb_last),
                                      GDK_PAINTABLE(args->thumb));
@@ -324,33 +413,32 @@ static void
 position_preview(float *offset_x, float *offset_y, float *size_x, float *size_y)
 {
         int buffer_width, buffer_height;
-        if (device_rotation == 0 || device_rotation == 180) {
-                buffer_width = preview_buffer_width;
-                buffer_height = preview_buffer_height;
+        if (state.device_rotation == 0 || state.device_rotation == 180) {
+                buffer_width = state.preview_buffer_width;
+                buffer_height = state.preview_buffer_height;
         } else {
-                buffer_width = preview_buffer_height;
-                buffer_height = preview_buffer_width;
+                buffer_width = state.preview_buffer_height;
+                buffer_height = state.preview_buffer_width;
         }
 
         int scale_factor = gtk_widget_get_scale_factor(preview);
-        int top_height =
-                gtk_widget_get_allocated_height(preview_top_box) * scale_factor;
-        int bottom_height =
-                gtk_widget_get_allocated_height(preview_bottom_box) * scale_factor;
-        int inner_height = preview_height - top_height - bottom_height;
+        int top_height = gtk_widget_get_height(preview_top_box) * scale_factor;
+        int bottom_height = gtk_widget_get_height(preview_bottom_box) * scale_factor;
+        int inner_height = state.preview_height - top_height - bottom_height;
 
-        double scale = MIN(preview_width / (float)buffer_width,
-                           preview_height / (float)buffer_height);
+        float scale = (float)MIN(state.preview_width / (float)buffer_width,
+                                 state.preview_height / (float)buffer_height);
 
-        *size_x = scale * buffer_width;
-        *size_y = scale * buffer_height;
+        *size_x = scale * (float)buffer_width;
+        *size_y = scale * (float)buffer_height;
 
-        *offset_x = (preview_width - *size_x) / 2.0;
+        *offset_x = ((float)state.preview_width - *size_x) / 2.0f;
 
-        if (*size_y > inner_height) {
-                *offset_y = (preview_height - *size_y) / 2.0;
+        if (*size_y > (float)inner_height) {
+                *offset_y = ((float)state.preview_height - *size_y) / 2.0f;
         } else {
-                *offset_y = top_height + (inner_height - *size_y) / 2.0;
+                *offset_y =
+                        (float)top_height + ((float)inner_height - *size_y) / 2.0f;
         }
 }
 
@@ -361,7 +449,7 @@ preview_draw(GtkGLArea *area, GdkGLContext *ctx, gpointer data)
                 return FALSE;
         }
 
-        if (!camera_is_initialized) {
+        if (current_preview_buffer == NULL) {
                 return FALSE;
         }
 
@@ -376,13 +464,15 @@ preview_draw(GtkGLArea *area, GdkGLContext *ctx, gpointer data)
 
         float offset_x, offset_y, size_x, size_y;
         position_preview(&offset_x, &offset_y, &size_x, &size_y);
-        glViewport(offset_x, preview_height - size_y - offset_y, size_x, size_y);
+        glViewport(
+                offset_x, state.preview_height - size_y - offset_y, size_x, size_y);
 
+        // This rotates the camera preview based on the phone rotation
         if (current_preview_buffer) {
                 glUseProgram(blit_program);
 
                 GLfloat rotation_list[4] = { 0, -1, 0, 1 };
-                int rotation_index = device_rotation / 90;
+                int rotation_index = state.device_rotation / 90;
 
                 GLfloat sin_rot = rotation_list[rotation_index];
                 GLfloat cos_rot = rotation_list[(4 + rotation_index - 1) % 4];
@@ -425,19 +515,38 @@ preview_draw(GtkGLArea *area, GdkGLContext *ctx, gpointer data)
                         MPZBarCode *code = &zbar_result->codes[i];
 
                         GLfloat vertices[] = {
-                                code->bounds_x[0], code->bounds_y[0],
-                                code->bounds_x[1], code->bounds_y[1],
-                                code->bounds_x[3], code->bounds_y[3],
-                                code->bounds_x[2], code->bounds_y[2],
+                                code->bounds_x[0], code->bounds_y[0], // Bottom left
+                                code->bounds_x[1], code->bounds_y[1], // Bottom right
+                                code->bounds_x[3], code->bounds_y[3], // Top left
+                                code->bounds_x[2], code->bounds_y[2], // Top right
                         };
 
+                        // Convert the pixel X/Y coordinates to OpenGL coordinates
+                        // (-1 to 1)
                         for (int i = 0; i < 4; ++i) {
-                                vertices[i * 2] =
-                                        2 * vertices[i * 2] / preview_buffer_width -
-                                        1.0;
-                                vertices[i * 2 + 1] =
-                                        1.0 - 2 * vertices[i * 2 + 1] /
-                                                      preview_buffer_height;
+                                // Width/height need to be swapped between portrait
+                                // mode and landscape mode for symbols to render
+                                // correctly
+                                if (state.device_rotation == 0 ||
+                                    state.device_rotation == 180) {
+                                        vertices[i * 2] =
+                                                2 * vertices[i * 2] /
+                                                        state.preview_buffer_width -
+                                                1.0;
+                                        vertices[i * 2 + 1] =
+                                                1.0 -
+                                                2 * vertices[i * 2 + 1] /
+                                                        state.preview_buffer_height;
+                                } else {
+                                        vertices[i * 2] =
+                                                2 * vertices[i * 2] /
+                                                        state.preview_buffer_height -
+                                                1.0;
+                                        vertices[i * 2 + 1] =
+                                                1.0 -
+                                                2 * vertices[i * 2 + 1] /
+                                                        state.preview_buffer_width;
+                                }
                         }
 
                         if (gtk_gl_area_get_use_es(area)) {
@@ -489,9 +598,11 @@ preview_draw(GtkGLArea *area, GdkGLContext *ctx, gpointer data)
 static gboolean
 preview_resize(GtkWidget *widget, int width, int height, gpointer data)
 {
-        if (preview_width != width || preview_height != height) {
-                preview_width = width;
-                preview_height = height;
+        if (state.preview_width != width || state.preview_height != height) {
+                state.preview_width = width;
+                state.preview_height = height;
+                if (state.configuration->count == 0)
+                        return TRUE;
                 update_io_pipeline();
         }
 
@@ -507,7 +618,7 @@ run_open_last_action(GSimpleAction *action, GVariant *param, gpointer user_data)
         if (strlen(last_path) == 0) {
                 return;
         }
-        sprintf(uri, "file://%s", last_path);
+        snprintf(uri, sizeof(uri), "file://%s", last_path);
         if (!g_app_info_launch_default_for_uri(uri, NULL, &error)) {
                 g_printerr("Could not launch image viewer for '%s': %s\n",
                            uri,
@@ -520,7 +631,7 @@ run_open_photos_action(GSimpleAction *action, GVariant *param, gpointer user_dat
 {
         char uri[270];
         g_autoptr(GError) error = NULL;
-        sprintf(uri, "file://%s", g_get_user_special_dir(G_USER_DIRECTORY_PICTURES));
+        snprintf(uri, sizeof(uri), "file://%s", g_get_user_special_dir(G_USER_DIRECTORY_PICTURES));
         if (!g_app_info_launch_default_for_uri(uri, NULL, &error)) {
                 g_printerr("Could not launch image viewer: %s\n", error->message);
         }
@@ -547,11 +658,11 @@ run_about_action(GSimpleAction *action, GVariant *param, GApplication *app)
                               "title",
                               "Megapixels",
                               "logo-icon-name",
-                              "org.postmarketos.Megapixels",
+                              "me.gapixels.Megapixels",
                               "comments",
                               "The postmarketOS camera application",
                               "website",
-                              "https://gitlab.com/postmarketOS/megapixels",
+                              "https://gitlab.com/megapixels-org/Megapixels",
                               "version",
                               VERSION,
                               "license-type",
@@ -566,7 +677,7 @@ run_quit_action(GSimpleAction *action, GVariant *param, GApplication *app)
 }
 
 static bool
-check_point_inside_bounds(int x, int y, int *bounds_x, int *bounds_y)
+check_point_inside_bounds(int x, int y, const int *bounds_x, const int *bounds_y)
 {
         bool right = false, left = false, top = false, bottom = false;
 
@@ -650,7 +761,7 @@ on_zbar_code_tapped(GtkWidget *widget, const MPZBarCode *code)
         g_signal_connect(
                 dialog, "response", G_CALLBACK(on_zbar_dialog_response), data);
 
-        gtk_widget_show(GTK_WIDGET(dialog));
+        gtk_widget_set_visible(GTK_WIDGET(dialog), true);
 }
 
 static void
@@ -667,9 +778,9 @@ preview_pressed(GtkGestureClick *gesture, int n_press, double x, double y)
                 position_preview(&offset_x, &offset_y, &size_x, &size_y);
 
                 int zbar_x = (x - offset_x) * scale_factor / size_x *
-                             preview_buffer_width;
+                             state.preview_buffer_width;
                 int zbar_y = (y - offset_y) * scale_factor / size_y *
-                             preview_buffer_height;
+                             state.preview_buffer_height;
 
                 for (uint8_t i = 0; i < zbar_result->size; ++i) {
                         MPZBarCode *code = &zbar_result->codes[i];
@@ -685,7 +796,7 @@ preview_pressed(GtkGestureClick *gesture, int n_press, double x, double y)
         }
 
         // Tapped preview image itself, try focussing
-        if (has_auto_focus_start) {
+        if (state.has_auto_focus_start) {
                 mp_io_pipeline_focus();
         }
 }
@@ -693,16 +804,14 @@ preview_pressed(GtkGestureClick *gesture, int n_press, double x, double y)
 static void
 run_camera_switch_action(GSimpleAction *action, GVariant *param, gpointer user_data)
 {
-        size_t next_index = camera->index + 1;
-        const struct mp_camera_config *next_camera =
-                mp_get_camera_config(next_index);
-
-        if (!next_camera) {
-                next_index = 0;
-                next_camera = mp_get_camera_config(next_index);
+        int new_index = state.camera->index + 1;
+        if (new_index > state.configuration->count - 1 || new_index < 0) {
+                new_index = 0;
         }
 
-        camera = next_camera;
+        state.camera = state.configuration->cameras[new_index];
+        // TODO: allow setting burst length in the config
+        state.burst_length = 5;
         update_io_pipeline();
 }
 
@@ -797,8 +906,8 @@ open_controls(GtkWidget *parent,
 static void
 set_gain(double value)
 {
-        if (gain != (int)value) {
-                gain = value;
+        if (state.gain.value != (int)value) {
+                state.gain.value_req = (int)value;
                 update_io_pipeline();
         }
 }
@@ -806,8 +915,8 @@ set_gain(double value)
 static void
 set_gain_auto(bool is_auto)
 {
-        if (gain_is_manual != !is_auto) {
-                gain_is_manual = !is_auto;
+        if (state.gain.manual != !is_auto) {
+                state.gain.manual_req = !is_auto;
                 update_io_pipeline();
         }
 }
@@ -818,9 +927,9 @@ open_iso_controls(GtkWidget *button, gpointer user_data)
         open_controls(button,
                       "ISO",
                       0,
-                      gain_max,
-                      gain,
-                      !gain_is_manual,
+                      state.gain.max,
+                      state.gain.value,
+                      !state.gain.manual,
                       set_gain,
                       set_gain_auto);
 }
@@ -828,9 +937,9 @@ open_iso_controls(GtkWidget *button, gpointer user_data)
 static void
 set_shutter(double value)
 {
-        int new_exposure = (int)(value / 360.0 * camera->capture_mode.height);
-        if (new_exposure != exposure) {
-                exposure = new_exposure;
+        int new_exposure = (int)(value / 360.0 * state.camera->current_mode->height);
+        if (new_exposure != state.exposure.value) {
+                state.exposure.value_req = new_exposure;
                 update_io_pipeline();
         }
 }
@@ -838,8 +947,26 @@ set_shutter(double value)
 static void
 set_shutter_auto(bool is_auto)
 {
-        if (exposure_is_manual != !is_auto) {
-                exposure_is_manual = !is_auto;
+        if (state.exposure.manual != !is_auto) {
+                state.exposure.manual_req = !is_auto;
+                update_io_pipeline();
+        }
+}
+
+static void
+set_focus(double value)
+{
+        if (state.focus.value != (int)value) {
+                state.focus.value_req = (int)value;
+                update_io_pipeline();
+        }
+}
+
+static void
+set_focus_auto(bool is_auto)
+{
+        if (state.focus.manual != !is_auto) {
+                state.focus.manual_req = !is_auto;
                 update_io_pipeline();
         }
 }
@@ -847,34 +974,81 @@ set_shutter_auto(bool is_auto)
 static void
 open_shutter_controls(GtkWidget *button, gpointer user_data)
 {
+        float value = (float)state.exposure.value * 360.0f /
+                      (float)state.camera->current_mode->height;
         open_controls(button,
                       "Shutter",
                       1.0,
                       360.0,
-                      exposure,
-                      !exposure_is_manual,
+                      value,
+                      !state.exposure.manual,
                       set_shutter,
                       set_shutter_auto);
 }
 
+static void
+open_focus_controls(GtkWidget *button, gpointer user_data)
+{
+        open_controls(button,
+                      "Focus",
+                      0,
+                      state.focus.max,
+                      state.focus.value,
+                      !state.focus.manual,
+                      set_focus,
+                      set_focus_auto);
+}
+
 static void
 flash_button_clicked(GtkWidget *button, gpointer user_data)
 {
-        flash_enabled = !flash_enabled;
+        state.flash_enabled = !state.flash_enabled;
         update_io_pipeline();
 
-        const char *icon_name =
-                flash_enabled ? "flash-enabled-symbolic" : "flash-disabled-symbolic";
+        const char *icon_name = state.flash_enabled ? "flash-enabled-symbolic" :
+                                                      "flash-disabled-symbolic";
         gtk_button_set_icon_name(GTK_BUTTON(button), icon_name);
 }
 
+void
+notify_movie_record_ready(void)
+{
+	movie_start = 0;
+	gtk_button_set_icon_name(GTK_BUTTON(movie),
+				"video-inactive-symbolic");
+}
+
+void
+notify_movie_message(gchar *msg)
+{
+        gtk_button_set_label(GTK_BUTTON(movie), msg);
+}
+
+void
+on_movie_clicked(GtkWidget *widget, gpointer user_data)
+{
+        if (movie_start) {
+                movie_start = 0;
+                gtk_button_set_icon_name(GTK_BUTTON(movie),
+                                         "video-processing-symbolic");
+                on_movie_stop();
+        } else {
+                movie_start = 1;
+                gtk_button_set_icon_name(GTK_BUTTON(movie),
+                                         "video-recording-symbolic");
+                on_movie_start();
+        }
+}
+
 static void
 on_realize(GtkWidget *window, gpointer *data)
 {
+        if (state.configuration->count == 0) {
+                return;
+        }
         GtkNative *native = gtk_widget_get_native(window);
         mp_process_pipeline_init_gl(gtk_native_get_surface(native));
-
-        camera = mp_get_camera_config(0);
+        state.camera = state.configuration->cameras[0];
         update_io_pipeline();
 }
 
@@ -890,7 +1064,7 @@ create_simple_action(GtkApplication *app, const char *name, GCallback callback)
 static void
 update_ui_rotation()
 {
-        if (device_rotation == 0 || device_rotation == 180) {
+        if (state.device_rotation == 0 || state.device_rotation == 180) {
                 // Portrait
                 gtk_widget_set_halign(preview_top_box, GTK_ALIGN_FILL);
                 gtk_orientable_set_orientation(GTK_ORIENTABLE(preview_top_box),
@@ -900,7 +1074,7 @@ update_ui_rotation()
                 gtk_orientable_set_orientation(GTK_ORIENTABLE(preview_bottom_box),
                                                GTK_ORIENTATION_HORIZONTAL);
 
-                if (device_rotation == 0) {
+                if (state.device_rotation == 0) {
                         gtk_widget_set_valign(preview_top_box, GTK_ALIGN_START);
                         gtk_widget_set_valign(preview_bottom_box, GTK_ALIGN_END);
                 } else {
@@ -917,7 +1091,7 @@ update_ui_rotation()
                 gtk_orientable_set_orientation(GTK_ORIENTABLE(preview_bottom_box),
                                                GTK_ORIENTATION_VERTICAL);
 
-                if (device_rotation == 90) {
+                if (state.device_rotation == 90) {
                         gtk_widget_set_halign(preview_top_box, GTK_ALIGN_END);
                         gtk_widget_set_halign(preview_bottom_box, GTK_ALIGN_START);
                 } else {
@@ -961,14 +1135,14 @@ static gboolean
 fb_profile_to_state(GValue *value, GVariant *variant, gpointer user_data)
 {
         const gchar *name;
-        gboolean state = FALSE;
+        gboolean fb_state = FALSE;
 
         name = g_variant_get_string(variant, NULL);
 
         if (g_strcmp0(name, "full") == 0)
-                state = TRUE;
+                fb_state = TRUE;
 
-        g_value_set_boolean(value, state);
+        g_value_set_boolean(value, fb_state);
 
         return TRUE;
 }
@@ -978,9 +1152,9 @@ state_to_fb_profile(const GValue *value,
                     const GVariantType *expected_type,
                     gpointer user_data)
 {
-        gboolean state = g_value_get_boolean(value);
+        gboolean fb_state = g_value_get_boolean(value);
 
-        return g_variant_new_string(state ? "full" : "silent");
+        return g_variant_new_string(fb_state ? "full" : "silent");
 }
 
 static void
@@ -1033,8 +1207,8 @@ wl_handle_geometry(void *data,
         assert(transform < 4);
         int new_rotation = transform * 90;
 
-        if (new_rotation != device_rotation) {
-                device_rotation = new_rotation;
+        if (new_rotation != state.device_rotation) {
+                state.device_rotation = new_rotation;
                 update_io_pipeline();
                 update_ui_rotation();
         }
@@ -1107,8 +1281,8 @@ xevent_handler(GdkDisplay *display, XEvent *xevent, gpointer data)
                         new_rotation = 270;
                         break;
                 }
-                if (new_rotation != device_rotation) {
-                        device_rotation = new_rotation;
+                if (new_rotation != state.device_rotation) {
+                        state.device_rotation = new_rotation;
                         update_io_pipeline();
                         update_ui_rotation();
                 }
@@ -1120,6 +1294,17 @@ xevent_handler(GdkDisplay *display, XEvent *xevent, gpointer data)
 }
 #endif // GDK_WINDOWING_X11
 
+static void
+rotation_sensor_handler(int rotation, bool good)
+{
+         if (rotation != state.device_accel_rotation ||
+             good != state.device_accel_rotation_good) {
+                 state.device_accel_rotation = rotation;
+                 state.device_accel_rotation_good = good;
+                 update_io_pipeline();
+         }
+}
+
 static void
 activate(GtkApplication *app, gpointer data)
 {
@@ -1142,13 +1327,17 @@ activate(GtkApplication *app, gpointer data)
         GtkBuilder *builder = gtk_builder_new_from_resource(
                 "/org/postmarketos/Megapixels/camera.ui");
 
-        GtkWidget *window = GTK_WIDGET(gtk_builder_get_object(builder, "window"));
-        GtkWidget *iso_button =
+        window = GTK_WIDGET(gtk_builder_get_object(builder, "window"));
+        iso_button =
                 GTK_WIDGET(gtk_builder_get_object(builder, "iso-controls-button"));
-        GtkWidget *shutter_button = GTK_WIDGET(
+        shutter_button = GTK_WIDGET(
                 gtk_builder_get_object(builder, "shutter-controls-button"));
+        focus_button =
+                GTK_WIDGET(gtk_builder_get_object(builder, "focus-controls-button"));
         flash_button =
                 GTK_WIDGET(gtk_builder_get_object(builder, "flash-controls-button"));
+        status_button =
+                GTK_WIDGET(gtk_builder_get_object(builder, "status-button"));
         GtkWidget *setting_dng_button =
                 GTK_WIDGET(gtk_builder_get_object(builder, "setting-raw"));
         GtkWidget *setting_postprocessor_combo =
@@ -1166,6 +1355,10 @@ activate(GtkApplication *app, gpointer data)
         preview_top_box = GTK_WIDGET(gtk_builder_get_object(builder, "top-box"));
         preview_bottom_box =
                 GTK_WIDGET(gtk_builder_get_object(builder, "bottom-box"));
+        movie = GTK_WIDGET(gtk_builder_get_object(builder, "movie"));
+
+        message_box = GTK_WIDGET(gtk_builder_get_object(builder, "message-box"));
+        message_label = GTK_WIDGET(gtk_builder_get_object(builder, "message-label"));
 
         g_signal_connect(window, "realize", G_CALLBACK(on_realize), NULL);
 
@@ -1179,9 +1372,13 @@ activate(GtkApplication *app, gpointer data)
         g_signal_connect(iso_button, "clicked", G_CALLBACK(open_iso_controls), NULL);
         g_signal_connect(
                 shutter_button, "clicked", G_CALLBACK(open_shutter_controls), NULL);
+        g_signal_connect(
+                focus_button, "clicked", G_CALLBACK(open_focus_controls), NULL);
         g_signal_connect(
                 flash_button, "clicked", G_CALLBACK(flash_button_clicked), NULL);
 
+        g_signal_connect(movie, "clicked", G_CALLBACK(on_movie_clicked), NULL);
+
         setup_fb_switch(builder);
 
         // Setup actions
@@ -1198,28 +1395,35 @@ activate(GtkApplication *app, gpointer data)
         create_simple_action(app, "quit", G_CALLBACK(run_quit_action));
 
         // Setup shortcuts
-        const char *capture_accels[] = { "space", NULL };
+        const char *capture_accels[] = {
+                "space",
+                "AudioLowerVolume",
+                "AudioRaiseVolume",
+                NULL
+        };
         gtk_application_set_accels_for_action(app, "app.capture", capture_accels);
 
         const char *quit_accels[] = { "<Ctrl>q", "<Ctrl>w", NULL };
         gtk_application_set_accels_for_action(app, "app.quit", quit_accels);
 
         // Setup settings
-        settings = g_settings_new("org.postmarketos.Megapixels");
+        settings = g_settings_new(APP_ID);
         char *setting_postproc = g_settings_get_string(settings, "postprocessor");
 
         // Initialize the postprocessing gsetting to the old processor if
         // it was not set yet
         if (setting_postproc == NULL || setting_postproc[0] == '\0') {
+		const int size = 512;
                 printf("Initializing postprocessor gsetting\n");
-                setting_postproc = malloc(512);
-                if (!mp_process_find_processor(setting_postproc)) {
+                setting_postproc = malloc(size);
+                if (!mp_process_find_processor(setting_postproc, size, "postprocess.sh")) {
                         printf("No processor found\n");
                         exit(1);
                 }
                 g_settings_set_string(settings, "postprocessor", setting_postproc);
                 printf("Initialized postprocessor to %s\n", setting_postproc);
         }
+	free(setting_postproc);
 
         // Find all postprocessors for the settings list
         mp_process_find_all_processors(setting_postprocessor_list);
@@ -1278,8 +1482,8 @@ activate(GtkApplication *app, gpointer data)
                         new_rotation = 270;
                         break;
                 }
-                if (new_rotation != device_rotation) {
-                        device_rotation = new_rotation;
+                if (new_rotation != state.device_rotation) {
+                        state.device_rotation = new_rotation;
                         update_ui_rotation();
                 }
         }
@@ -1289,11 +1493,22 @@ activate(GtkApplication *app, gpointer data)
         GDBusConnection *conn =
                 g_application_get_dbus_connection(G_APPLICATION(app));
         mp_flash_gtk_init(conn);
+        mp_device_rotation_init(rotation_sensor_handler);
 
-        mp_io_pipeline_start();
+        if (state.configuration->count > 0) {
+                mp_io_pipeline_start();
+        } else {
+                display_error("No camera found");
+        }
 
         gtk_application_add_window(app, GTK_WINDOW(window));
-        gtk_widget_show(window);
+        gtk_widget_set_visible(window, true);
+
+        // Inhibit system shortcuts to be able to use
+        // VolumeUp and VolumeDown hardware buttons to take a picture
+        GtkNative *native = gtk_widget_get_native (GTK_WIDGET (window));
+        GdkSurface *surface = gtk_native_get_surface (native);
+        gdk_toplevel_inhibit_system_shortcuts (GDK_TOPLEVEL (surface), NULL);
 }
 
 static void
@@ -1314,6 +1529,7 @@ shutdown(GApplication *app, gpointer data)
 #ifdef DEBUG
         mp_io_pipeline_stop();
         mp_flash_gtk_clean();
+        mp_device_rotation_clean();
 
         g_clear_object(&fb_settings);
         g_clear_object(&capture_event);
@@ -1338,9 +1554,20 @@ main(int argc, char *argv[])
                 }
         }
 #endif
-
-        if (!mp_load_config())
-                return 1;
+        char configfile[PATH_MAX];
+        libmegapixels_init(&state.configuration);
+        if (libmegapixels_find_config(PATH_MAX, configfile)) {
+                if (!libmegapixels_load_file(state.configuration, configfile)) {
+                        fprintf(stderr, "Could not load config\n");
+                        return 1;
+                }
+        } else {
+                printf("megapixels: No suitable config, defaulting to uvc\n");
+                if (!libmegapixels_load_uvc(state.configuration)) {
+                        fprintf(stderr, "No config found\n");
+                        return 1;
+                }
+        }
 
         setenv("LC_NUMERIC", "C", 1);
 
