@@ -1,14 +1,14 @@
 #pragma once
 
 #include "camera.h"
-#include "camera_config.h"
+#include "state.h"
 #include <gtk/gtk.h>
 
 typedef struct _GdkSurface GdkSurface;
 
 struct mp_process_pipeline_state {
-        const struct mp_camera_config *camera;
-        MPMode mode;
+        libmegapixels_camera *camera;
+        libmegapixels_devconfig *configuration;
 
         int burst_length;
 
@@ -31,9 +31,14 @@ struct mp_process_pipeline_state {
         bool has_auto_focus_start;
 
         bool flash_enabled;
+
+        bool control_gain;
+        bool control_exposure;
+        bool control_focus;
 };
 
-bool mp_process_find_processor(char *script);
+
+bool mp_process_find_processor(char *script, int size, char *filename);
 void mp_process_find_all_processors(GtkListStore *store);
 
 void mp_process_pipeline_start();
@@ -44,7 +49,10 @@ void mp_process_pipeline_init_gl(GdkSurface *window);
 
 void mp_process_pipeline_process_image(MPBuffer buffer);
 void mp_process_pipeline_capture();
-void mp_process_pipeline_update_state(const struct mp_process_pipeline_state *state);
+void mp_process_pipeline_update_state(const mp_state_proc *new_state);
+
+void on_movie_start(void);
+void on_movie_stop(void);
 
 typedef struct _MPProcessPipelineBuffer MPProcessPipelineBuffer;
 
